@testset "solve" begin
    @testset "Invalid input" begin
        @polyvar x y z
        @test_throws ErrorException solve([x-2y+2, 0])
        @test_throws ErrorException solve([x-2, y-2], [x-2, y-2,y+2], [[2, -3]])
        @test_throws ErrorException solve([x-2z, y^2+3z, z^3+x^3], homvar=z)
        # overdetermined and abstract system
        @test_throws ErrorException solve(FPSystem([x-2z, y^2+3z^2, z^3+x^3, z+x]))
        @test_throws ErrorException solve(FPSystem([x-2z, y^2+3z^2, z^3+x^3, z+x]), homvar=4)

        @test_throws ErrorException solve([x-2z, y^2+3z^2, z^3+x^3])

        @test_throws ErrorException solve([x-2z, y^2+3z], homvar=z)

        # invalid kwargs
        @test_throws ErrorException solve(equations(cyclic(5)), def=0.4, abc=23)

        # test numerical homogeneous check fails
        @polyvar x y z
        G = FPSystem([x-2z, y^2+3z])
        @test_throws ErrorException solve(G, homvar=3)
    end

    @testset "solve" begin
        @polyvar x
        @test nfinite(solve([x - 1])) == 1
        F = equations(katsura(5))
        @test nfinite(solve(F, threading=false)) == 32
        @test nfinite(solve(F, system=SPSystem, threading=false)) == 32
        @test nfinite(solve(F, system=FPSystem, threading=false)) == 32

        # Simple step size
        F = equations(katsura(5))
        @test nfinite(solve(F, simple_step_size_alg=true, threading=false)) == 32

        # scaling
        F = equations(katsura(5))
        @test nfinite(solve(F, system_scaling=false, threading=false)) == 32

        @test nfinite(solve(F, homotopy=StraightLineHomotopy)) == 32
        result = solve(F, predictor=Euler(), homotopy=StraightLineHomotopy)
        @test nresults(result) == 32
        @test nfinite(solve(F, accuracy=1e-5)) == 32

        result = solve(F)
        @test nfinite(result) == 32
        @test string(result) isa String

        @test nfinite(solve(F, patch=RandomPatch())) == 32
        @test nfinite(solve(F, patch=EmbeddingPatch())) ≤ 32
        @test nfinite(solve(F, patch=OrthogonalPatch())) == 32

        @polyvar w
        F = equations(cyclic(5))
        result = solve(homogenize(F, w), threading=false, homvar=w)
        @test result isa Result

        @polyvar x y z

        G = [x-2, y+3]
        F = [x+2, y-3]
        @test nfinite(solve(G, F, [[2, -3]])) == 1
        @test nfinite(solve(G, F, [[2+0.0im, -3.0+0im]])) == 1

        @test nfinite(solve(G, F, [@SVector [2, -3]])) == 1

        F = FPSystem(homogenize(equations(cyclic(5))))
        result = solve(F, homvar=6, save_all_paths=true)
        @test nfinite(result) == 70
        @test natinfinity(result) == 50


        F = equations(katsura(5))
        result = solve(homogenize(F))
        @test result isa Result{<:ProjectiveVectors.PVector}
        @test isprojective(first(result))
        @test nnonsingular(result) == 32


        prob, startsolutions = problem_startsolutions(TotalDegreeInput(F))
        result = solve(prob.homotopy, map(startsolutions) do s
            HC.embed(prob, s)
        end)
        @test result isa Result{<:ProjectiveVectors.PVector}
        @test nnonsingular(result) == 32

        result = solve(prob.homotopy, map(startsolutions) do s
            HC.embed(prob, s)
        end, homvar=homvars(prob)[1])
        @test result isa Result{<:Vector}
        @test nnonsingular(result) == 32
        @test nfinite(result) == 32

        # Composition
        @polyvar a b c x y z u v
        e = [u + 1, v - 2]
        f = [a * b - 2, a*c- 1]
        g = [x+y, y + 3, x + 2]
        res = solve(e ∘ f ∘ g)
        @test nnonsingular(res) == 2
        @test nnonsingular(solve(e ∘ f ∘ g, system_scaling=false)) == 2

        res = solve(e ∘ f ∘ g, system=SPSystem)
        @test nnonsingular(res) == 2
    end


    @testset "Random seed" begin
        R = solve(equations(katsura(5)), seed=1234)
        @test seed(R) == 1234

        @polyvar x y

        G = [x-2, y+3]
        F = [x+2, y-3]
        @test seed(solve(G, F, [[2, -3]], seed=222)) == 222
    end

    @testset "Singular solutions" begin
        @polyvar x y z
        z = 1
        F = [x^2 + 2*y^2 + 2*im*y*z, (18 + 3*im)*x*y + 7*im*y^2 - (3 - 18*im)*x*z - 14*y*z - 7*im*z^2]
        result = solve(F)
        @test nsingular(result) == 3
        @test all(r -> r.winding_number == 3, singular(result))
    end

    @testset "Path jumping" begin
        result = solve(equations(katsura(5)); accuracy=1e-2, refinement_accuracy=1e-8, seed=39813, threading=false)
        @test nreal(result) == 16
    end

    @testset "Affine vs projective" begin
        @polyvar x y z
        f = [x-2y, y-2z]
        g = [x-2, y-2]

        F = solve(f)
        G = solve(g)

        @test length(F[1].solution) == 3
        @test length(G[1].solution) == 2
        @test F[1].solution isa ProjectiveVectors.PVector
        @test G[1].solution isa Vector
    end

    @testset "Parameter Homotopies" begin
        @polyvar x a y b
        F = [x^2-a, x*y-a+b]
        p = [a, b]
        S = solve(F, [[1.0, 1.0 + 0.0*im]], parameters=p, p₁=[1, 0], p₀=[2, 4])
        # S = solve(F, p, [1, 0], [2, 4], [[1.0, 1.0 + 0.0*im]])

        @test S[1].solution ≈ [complex(√2), -complex(√2)]
        @test nfinite(S) == 1

        @polyvar x a y b z
        F = [x^2-a*z^2, x*y-(a-b)*z^2]
        S = solve(F, [[1.0, 1.0 + 0.0*im, 1.0]], parameters=[a, b], start_parameters=[1, 0], target_parameters=[2, 4])
        @test S isa Result{<:ProjectiveVectors.PVector}
        @test solution(S[1])[1:2] / solution(S[1])[3] ≈ [complex(√2), -complex(√2)]
        @test nnonsingular(S) == 1

        S2 = solve(F, [[1.0, 1.0 + 0.0*im, 1.0]], parameters=[a, b], p₁=[1, 0], p₀=[2, 4], homvar=z)
        @test solution(S2[1]) ≈ [complex(√2), -complex(√2)]
        @test nfinite(S2) == 1

        γ₁, γ₀ =randn(ComplexF64, 2)
        S2 = solve(F, [[1.0, 1.0 + 0.0*im, 1.0]], parameters=[a, b], p₁=[1, 0], p₀=[2, 4], γ₁=γ₁, γ₀=γ₀, homvar=z)
        @test solution(S2[1]) ≈ [complex(√2), -complex(√2)]
        @test nfinite(S2) == 1


        γ₁, γ₀ =randn(ComplexF64, 2)
        S2 = solve(F, [@SVector [1.0, 1.0 + 0.0*im, 1.0]], parameters=[a, b], p₁=[1, 0], p₀=[2, 4], γ₁=γ₁, γ₀=γ₀, homvar=z)
        @test solution(S2[1]) ≈ [complex(√2), -complex(√2)]
        @test nfinite(S2) == 1
    end

    @testset "ParameterHomotopy with Composition" begin
        @polyvar p q a b c x y z u v
        f = [a * b - 2, a*c- 1]
        g = [x + y, y + 3, x + 2]
        res = solve(f ∘ g, system=SPSystem)

        # parameters at the end
        f2 = [a * b - q, a*c- p]
        g = [x + y, y + 3, x + 2]
        r = solve(f2 ∘ g, solutions(res), parameters=[p, q], p₁=[1, 2], p₀=[2, 3])
        @test HomotopyContinuation.nnonsingular(r) == 2

        # parameters at the beginning
        f = [a * b - 2, a*c- 1]
        g2 = [x + y, y + u, x + v]
        r = solve(f ∘ g2, solutions(res), parameters=[u, v], p₁=[3, 2], p₀=[-2, 3])
        @test HomotopyContinuation.nnonsingular(r) == 2

        # parameter in the middle
        e = [u + 1, v - 2]
        res2 = solve(e ∘ f ∘ g, system=SPSystem)
        f2 = [a * b - q, a * c- p]
        r = solve(e ∘ f2 ∘ g, solutions(res2), parameters=[p, q], p₁=[1, 2], p₀=[2, 3])
        @test HomotopyContinuation.nnonsingular(r) == 2
    end

    @testset "Overdetermined" begin
        Random.seed!(1234567)
        @polyvar x y z w
        a = [0.713865+0.345317im, 0.705182+0.455495im, 0.9815+0.922608im, 0.337617+0.508932im]

        f = [x*z-y^2, y*w-z^2, x*w-y*z]
        L₁ = [1, -1, 1, -1] ⋅ [x, y, z, w]
        L₂ = rand(ComplexF64, 4) ⋅ [x, y, z, w]
        S = solve([f; L₁], [f; L₂], [[1, 1, 1, 1]])
        @test nnonsingular(S) == 1

        f = [x*z-y^2, y-z^2, x-y*z]
        L₁ = [1, -1, 1, -1] ⋅ [x, y, z, 1]
        L₂ = rand(ComplexF64, 4) ⋅ [x, y, z, 1]
        S = solve([f; L₁], [f; L₂], [[1, 1, 1]])
        @test nfinite(S) == 1


        @polyvar x y z
        p₁ = (y - x^2) * (x^2+y^2+z^2 - 1) * (x - 0.5)
        p₂ = (z - x^3) * (x^2+y^2+z^2 - 1) * (y - 0.5)
        p₃ = (y - x^2) * (z - x^3) * (x^2+y^2+z^2 - 1) * (z - 0.5)
        L₁ = [rand(ComplexF64, 4) ⋅ [x, y, z, 1], rand(ComplexF64, 4) ⋅ [x, y, z, 1]]
        L₂ = [rand(ComplexF64, 4) ⋅ [x, y, z, 1], rand(ComplexF64, 4) ⋅ [x, y, z, 1]]

        s = first(solutions(solve([x^2+y^2+z^2 - 1; L₁])))
        tracker, starts = pathtracker_startsolutions([[p₁, p₂, p₃]; L₁], [[p₁, p₂, p₃]; L₂], s)

        @test track(tracker, s).return_code == :success
    end

    @testset "MultiHomogeneous" begin
        @polyvar x y u v

        f = [x*y - 6, x^2 - 5]
        S = solve(f, variable_groups=[(x,), (y,)])
        @test nnonsingular(S) == 2
        S = solve(f, variable_groups=[(x,), (y,)])
        @test nnonsingular(S) == 2
        g = [x*y - 6u*v, x^2 - u^2]
        S = solve(g, variable_groups=[(x,u), (y,v)], homvars=(u,v))
        @test nnonsingular(S) == 2


        # The 6-R inverse problem

        ## initialize the variables
        @polyvar z[1:6,1:3]
        p = [1, 1, 0]
        α = randn(5)
        a = randn(9)
        ## define the system of polynomials
        f = [z[i,:] ⋅ z[i,:] for i=2:5]
        g = [z[i,:] ⋅ z[i+1,:] for i=1:5]
        h = sum(a[i] .* (z[i,:] × z[i+1,:]) for i=1:3) +
            sum(a[i+4] .* z[i,:] for i=2:5)
        F′ = [f .- 1; g .- cos.(α); h .- p]
        ## assign values to z₁ and z₆
        F = [subs(f, z[1,:] => [1, 0, 0], z[6,:] => [1,0,0]) for f in F′]

        group1 = [[z[2,:]; z[4,:]], [z[3,:]; z[5,:]]]
        group2 = [z[2,:], z[4,:], z[3,:], z[5,:]]

        @test bezout_number(F) == 1024
        @test bezout_number(F, variable_groups=group1) == 320
        @test bezout_number(F, variable_groups=group2) == 576

        @test nnonsingular(solve(F, seed=1234)) == 16
        @test nnonsingular(solve(F; variable_groups=group1, seed=1234)) == 16
    end

    @testset "Affine TotalDegree" begin
        result = solve(equations(cyclic(5)), affine_tracking=true, seed=12314)
        @test nnonsingular(result) == 70
        @test nsingular(result) == 0
    end

    @testset "overdetermiend total degree" begin
        @polyvar x y z
        # This are the maximal minors of a 3 by 5 matrix
        f = [-(19/2)*x^6-(306679/75600)*x^5*y+(3566281/453600)*x^4*y^2+(2458541/113400)*x^3*y^3+(1057993/97200)*x^2*y^4-(4093/28350)*x*y^5-(701/675)*y^6+(156769/4704)*x^5*z+(2136401/201600)*x^4*y*z-(485549/317520)*x^3*y^2*z-(1228333069/38102400)*x^2*y^3*z-(444121/12960)*x*y^4*z-(12491/2016)*y^5*z+(5554901/1411200)*x^4*z^2-(1396343/453600)*x^3*y*z^2-(2815352089/50803200)*x^2*y^2*z^2-(6727757957/127008000)*x*y^3*z^2-(4014469/378000)*y^4*z^2+(21830581/658560)*x^3*z^3+(208824821/14817600)*x^2*y*z^3-(105675343/9525600)*x*y^2*z^3-(744971/158760)*y^3*z^3+(4234567/352800)*x^2*z^4+(26384189/2116800)*x*y*z^4-(153973/952560)*y^2*z^4+(1465123/176400)*x*z^5+(537751/238140)*y*z^5+(121/315)*z^6+(160717/37800)*x^5+(3742241/378000)*x^4*y+(28611931/1134000)*x^3*y^2+(61601/36450)*x^2*y^3-(17656943/1701000)*x*y^4-(20531/5400)*y^5-(894547/84672)*x^4*z-(2346079/470400)*x^3*y*z-(16933223/396900)*x^2*y^2*z-(553452937/13608000)*x*y^3*z-(206119/34020)*y^4*z-(848783/705600)*x^3*z^2-(61574861/2540160)*x^2*y*z^2-(24620545003/457228800)*x*y^2*z^2-(92970251/9072000)*y^3*z^2-(1267237/423360)*x^2*z^3-(474156533/114307200)*x*y*z^3-(30023869/6804000)*y^2*z^3+(186664/19845)*x*z^4+(3533393/3175200)*y*z^4+(2557/14175)*z^5-(433/540)*x^4+(3979387/1134000)*x^3*y+(176181763/8164800)*x^2*y^2-(23612089/3402000)*x*y^3-(1099939/97200)*y^4-(3776473/423360)*x^3*z-(57493058237/1143072000)*x^2*y*z-(5521110767/54432000)*x*y^2*z+(29454571/2041200)*y^3*z-(313762747/6804000)*x^2*z^2-(995824699/38102400)*x*y*z^2-(123893137/4665600)*y^2*z^2+(17925923/2381400)*x*z^3+(5874637/1587600)*y*z^3-(813/140)*z^4-(5369/48600)*x^3-(1596487/453600)*x^2*y-(419513/32400)*x*y^2-(113359/21600)*y^3-(87364253/2721600)*x^2*z-(246098857/10886400)*x*y*z+(104373149/3628800)*y^2*z-(63677/3024)*x*z^2-(583877/388800)*y*z^2-(7757/3150)*z^3+(14549/6480)*x^2-(517663/75600)*x*y+(150341/8640)*y^2-(10412771/302400)*x*z+(209327/181440)*y*z+(5506/945)*z^2-(211/60)*x-(517/480)*y+(299/30)*z+47/16, (2/3)*x^6+(220411/37800)*x^5*y-(494933/113400)*x^4*y^2+(4298527/453600)*x^3*y^3+(794509/226800)*x^2*y^4+(4049/7200)*x*y^5+(21/80)*y^6-(4987/1575)*x^5*z-(1746887/75600)*x^4*y*z+(21906547/1360800)*x^3*y^2*z+(6301643/2041200)*x^2*y^3*z+(43001477/2721600)*x*y^4*z+(47231/12960)*y^5*z-(1141/2400)*x^4*z^2+(23865827/12700800)*x^3*y*z^2-(17041681/1190700)*x^2*y^2*z^2+(55657337/23814000)*x*y^3*z^2+(306001/75600)*y^4*z^2+(4657139/470400)*x^3*z^3-(11497061/793800)*x^2*y*z^3+(3973489/6096384)*x*y^2*z^3+(4007189/1693440)*y^3*z^3+(389769/156800)*x^2*z^4+(98207447/19051200)*x*y*z^4+(165523/79380)*y^2*z^4+(1689101/352800)*x*z^5+(12493993/6350400)*y*z^5+(2327/3600)*z^6+(1129/450)*x^5-(8123971/1134000)*x^4*y+(5767/2268)*x^3*y^2+(30397/170100)*x^2*y^3+(745733/151200)*x*y^4+(6509/2400)*y^5-(63893/8400)*x^4*z+(7910717/3175200)*x^3*y*z-(312932947/9525600)*x^2*y^2*z+(126425483/9072000)*x*y^3*z+(7200563/1360800)*y^4*z-(33705709/2116800)*x^3*z^2-(729523649/12700800)*x^2*y*z^2-(1374377629/47628000)*x*y^2*z^2+(28807441/28576800)*y^3*z^2-(52739317/6350400)*x^2*z^3-(34461347/88905600)*x*y*z^3+(64953487/91445760)*y^2*z^3-(10755347/6350400)*x*z^4+(6332959/2540160)*y*z^4+(161069/529200)*z^5+(317/108)*x^4+(36225143/1360800)*x^3*y-(737491/129600)*x^2*y^2+(2872463/259200)*x*y^3+(270253/129600)*y^4-(1155239/64800)*x^3*z-(91503641/3175200)*x^2*y*z+(19841081/1270080)*x*y^2*z-(18974807/5443200)*y^3*z-(43379321/6350400)*x^2*z^2-(6952763359/381024000)*x*y*z^2-(319039243/38102400)*y^2*z^2-(21814313/2646000)*x*z^3-(497789/5443200)*y*z^3-(25273/132300)*z^4+(193469/10800)*x^3-(12734041/2268000)*x^2*y+(264791/64800)*x*y^2+(525217/86400)*y^3+(279821/31500)*x^2*z+(108020863/10584000)*x*y*z-(19662193/907200)*y^2*z+(19939433/3528000)*x*z^2-(108970333/25401600)*y*z^2-(32677/70560)*z^3+(4453/10800)*x^2+(2576759/100800)*x*y-(27121/2880)*y^2+(59197/1400)*x*z+(1015583/302400)*y*z-(3773701/302400)*z^2+(15217/720)*x+(1379/960)*y+(15359/5040)*z+223/16, (29/7)*x^6+(23227/2520)*x^5*y+(917771/63000)*x^4*y^2+(11548/945)*x^3*y^3-(515479/14400)*x^2*y^4-(25571/2592)*x*y^5+(497/1080)*y^6-(130651/8400)*x^5*z-(21666527/504000)*x^4*y*z-(6470071/1296000)*x^3*y^2*z-(4645133/403200)*x^2*y^3*z+(18227527/259200)*x*y^4*z+(26387/1440)*y^5*z+(456553/235200)*x^4*z^2-(178328503/10584000)*x^3*y*z^2+(5343791/846720)*x^2*y^2*z^2+(114270983/1360800)*x*y^3*z^2+(12508007/453600)*y^4*z^2-(778381/78400)*x^3*z^3-(42594247/2646000)*x^2*y*z^3+(19687831/1036800)*x*y^2*z^3+(178310191/10886400)*y^3*z^3-(34949/7350)*x^2*z^4-(11336141/756000)*x*y*z^4-(1164427/1701000)*y^2*z^4-(5001/800)*x*z^5-(1089971/302400)*y*z^5-(1039/1050)*z^6+(3533/6300)*x^5-(1921583/189000)*x^4*y-(1492481/113400)*x^3*y^2-(180587/6048)*x^2*y^3+(966361/194400)*x*y^4+(10747/2592)*y^5-(11137493/352800)*x^4*z-(641388757/15876000)*x^3*y*z+(3712801/907200)*x^2*y^2*z+(601009153/4665600)*x*y^3*z+(36446017/907200)*y^4*z-(213277/14112)*x^3*z^2-(40330637/1134000)*x^2*y*z^2+(21871050209/571536000)*x*y^2*z^2+(29266459/907200)*y^3*z^2-(236459003/24696000)*x^2*z^3-(283084717/21168000)*x*y*z^3+(681321199/254016000)*y^2*z^3-(325699/117600)*x*z^4-(151170469/31752000)*y*z^4-(49801/110250)*z^5+(100679/3780)*x^4+(10379/1296)*x^3*y-(21673403/972000)*x^2*y^2+(4933739/194400)*x*y^3-(684991/25920)*y^4+(1054363/42336)*x^3*z+(442296401/190512000)*x^2*y*z+(727316399/27216000)*x*y^2*z+(30058829/907200)*y^3*z+(8978639/264600)*x^2*z^2+(1330320163/19051200)*x*y*z^2+(6313/4032)*y^2*z^2+(210615067/10584000)*x*z^3+(10294511/3024000)*y*z^3+(1093333/441000)*z^4+(77383/9450)*x^3+(507847/40500)*x^2*y+(52943/10800)*x*y^2+(92641/12960)*y^3+(60435509/5292000)*x^2*z+(517865647/9072000)*x*y*z+(2229677/907200)*y^2*z+(142591051/5292000)*x*z^2+(4027/960)*y*z^2+(26357/22050)*z^3+(126251/9450)*x^2+(17051/2700)*x*y-(134689/1440)*y^2-(770923/151200)*x*z+(913747/75600)*y*z+(12853/2800)*z^2-(1999/360)*x+(172/45)*y-(904/105)*z-209/8, -(10/3)*x^6-(1153/504)*x^5*y+(889/60)*x^4*y^2+(2360791/151200)*x^3*y^3+(10061/27000)*x^2*y^4-(509737/25200)*x*y^5-(5917/450)*y^6+(105163/8820)*x^5*z-(9789209/2116800)*x^4*y*z-(35670337/4233600)*x^3*y^2*z-(33335147/2646000)*x^2*y^3*z-(9131057/378000)*x*y^4*z+(139639/25200)*y^5*z+(192649/117600)*x^4*z^2+(8835811/846720)*x^3*y*z^2-(35411371/1587600)*x^2*y^2*z^2-(48672359/21168000)*x*y^3*z^2+(80051/10800)*y^4*z^2+(356102/33075)*x^3*z^3-(8227481/3628800)*x^2*y*z^3+(63073973/15240960)*x*y^2*z^3-(197/75600)*y^3*z^3-(1484633/1058400)*x^2*z^4+(9985471/1016064)*x*y*z^4-(12687223/1814400)*y^2*z^4+(150233/75600)*x*z^5-(166951/37800)*y*z^5-(20471/10800)*z^6-(1852/945)*x^5-(1189/11340)*x^4*y+(839971/75600)*x^3*y^2+(2531383/324000)*x^2*y^3-(329003/25200)*x*y^4-(227/20)*y^5-(419001/39200)*x^4*z-(274927927/13608000)*x^3*y*z-(294227203/28576800)*x^2*y^2*z+(46111091/3528000)*x*y^3*z+(21033239/453600)*y^4*z+(26793971/11113200)*x^3*z^2-(1884015269/111132000)*x^2*y*z^2+(180244423/47628000)*x*y^2*z^2+(27223717/1270080)*y^3*z^2+(2575621/3175200)*x^2*z^3+(38388299/2540160)*x*y*z^3+(41228297/4762800)*y^2*z^3+(413839/105840)*x*z^4-(13710967/1411200)*y*z^4-(154151/75600)*z^5+(2465143/113400)*x^4+(314857/162000)*x^3*y+(43081193/6804000)*x^2*y^2-(33754031/907200)*x*y^3-(507833/16200)*y^4-(1463879/198450)*x^3*z-(165791273/7938000)*x^2*y*z-(369287183/6804000)*x*y^2*z+(107246053/2721600)*y^3*z+(55352327/4762800)*x^2*z^2+(718293743/63504000)*x*y*z^2+(3794783/226800)*y^2*z^2+(26560811/3175200)*x*z^3+(100644931/12700800)*y*z^3+(266431/75600)*z^4-(38197/17010)*x^3+(3512851/1134000)*x^2*y-(63289/1680)*x*y^2+(1841/150)*y^3-(18876703/1587600)*x^2*z-(586122937/47628000)*x*y*z+(1279709/453600)*y^2*z-(3089503/1270080)*x*z^2+(15865867/1058400)*y*z^2+(40233/5600)*z^3-(256829/56700)*x^2-(465601/21600)*x*y-(82267/1440)*y^2-(1713977/50400)*x*z+(3229721/453600)*y*z+(7303/2240)*z^2-(13469/720)*x+(383/240)*y-(11999/1008)*z-285/16, -(489/40)*x^6-(77719/15120)*x^5*y+(267929/31752)*x^4*y^2+(14208101/1587600)*x^3*y^3+(3097037/793800)*x^2*y^4-(2249/9450)*x*y^5-(83/450)*y^6+(11783/315)*x^5*z-(380857/30240)*x^4*y*z+(326371/21600)*x^3*y^2*z-(2740261/635040)*x^2*y^3*z-(731603/105840)*x*y^4*z-(139/600)*y^5*z+(156029/10080)*x^4*z^2+(19124683/635040)*x^3*y*z^2-(226728533/88905600)*x^2*y^2*z^2+(39730727/9525600)*x*y^3*z^2+(5820139/2268000)*y^4*z^2+(3469769/66150)*x^3*z^3+(17835733/1190700)*x^2*y*z^3-(461024351/88905600)*x*y^2*z^3+(13947103/6350400)*y^3*z^3+(19490131/1058400)*x^2*z^4+(52893409/2381400)*x*y*z^4+(357967/119070)*y^2*z^4+(3239909/264600)*x*z^5+(22693901/4762800)*y*z^5+(187/140)*z^6+(11333/2700)*x^5+(12163517/680400)*x^4*y+(1459812847/71442000)*x^3*y^2+(373783/45360)*x^2*y^3-(39203/12960)*x*y^4-(23771/18000)*y^5-(114607/4725)*x^4*z-(13299883/302400)*x^3*y*z-(276735289/4762800)*x^2*y^2*z-(14432231/362880)*x*y^3*z-(19137691/4536000)*y^4*z-(380501/21600)*x^3*z^2-(260695451/3175200)*x^2*y*z^2-(630612179/9525600)*x*y^2*z^2-(216511/18144)*y^3*z^2-(1435591/635040)*x^2*z^3-(121932491/4762800)*x*y*z^3-(58163557/12700800)*y^2*z^3+(2120249/529200)*x*z^4+(267097/529200)*y*z^4+(2167/3240)*z^5+(11459/12960)*x^4+(1289/189)*x^3*y+(25911149/2381400)*x^2*y^2-(47656529/2721600)*x*y^3-(17533/2700)*y^4-(29615893/453600)*x^3*z-(1438245331/13608000)*x^2*y*z-(13010618809/142884000)*x*y^2*z-(5065363/1814400)*y^3*z-(112568063/2268000)*x^2*z^2-(14883632387/142884000)*x*y*z^2-(15060991/362880)*y^2*z^2-(20820743/1270080)*x*z^3-(14009717/2116800)*y*z^3-(1723/1512)*z^4+(93439/32400)*x^3+(350801/100800)*x^2*y-(15809887/604800)*x*y^2-(535/48)*y^3-(185643803/2721600)*x^2*z-(45604103/403200)*x*y*z+(1052449/201600)*y^2*z-(13916659/272160)*x*z^2-(10998347/302400)*y*z^2-(81359/10080)*z^3+(1457/4320)*x^2-(21497929/1209600)*x*y-(2233/240)*y^2-(54671779/907200)*x*z+(1877693/80640)*y*z-(6997/504)*z^2-(1701/320)*x+(9867/1280)*y+(7409/320)*z+837/64, -(561/140)*x^6+(460223/58800)*x^5*y-(7650443/352800)*x^4*y^2-(1992937/98000)*x^3*y^3-(39587/7875)*x^2*y^4-(28771/11340)*x*y^5+(301/675)*y^6-(37523/3360)*x^5*z-(38302421/4233600)*x^4*y*z+(55466501/705600)*x^3*y^2*z-(23967241/3175200)*x^2*y^3*z+(708251/113400)*x*y^4*z+(297571/37800)*y^5*z+(907777/117600)*x^4*z^2+(396988313/14817600)*x^3*y*z^2+(4007575717/63504000)*x^2*y^2*z^2+(202935527/1587600)*x*y^3*z^2+(15421667/453600)*y^4*z^2-(1030669/70560)*x^3*z^3+(150426089/74088000)*x^2*y*z^3+(506748581/9072000)*x*y^2*z^3+(295785457/10584000)*y^3*z^3-(2894573/352800)*x^2*z^4-(1558981/117600)*x*y*z^4+(373682/99225)*y^2*z^4-(52453/6300)*x*z^5-(2875571/529200)*y*z^5-(176/105)*z^6-(55739/6300)*x^5+(1501811/396900)*x^4*y+(19103233/756000)*x^3*y^2+(105225241/4536000)*x^2*y^3+(973957/272160)*x*y^4+(40121/16200)*y^5-(5148797/151200)*x^4*z-(139394203/1411200)*x^3*y*z-(4800962771/381024000)*x^2*y^2*z+(35899531/4536000)*x*y^3*z+(12601129/907200)*y^4*z-(1605559/529200)*x^3*z^2-(28872007/992250)*x^2*y*z^2+(152989777/10584000)*x*y^2*z^2+(33709733/1360800)*y^3*z^2-(16073177/4233600)*x^2*z^3-(1299883783/190512000)*x*y*z^3+(53429123/27216000)*y^2*z^3-(12347/31500)*x*z^4-(828467/588000)*y*z^4-(862/1575)*z^5+(90577/15120)*x^4+(15040213/6350400)*x^3*y+(86912731/3265920)*x^2*y^2-(72777953/2721600)*x*y^3+(115429/21600)*y^4-(13141267/604800)*x^3*z-(138795047659/2286144000)*x^2*y*z-(1068020413/27216000)*x*y^2*z+(43272007/1814400)*y^3*z+(301622123/63504000)*x^2*z^2+(574755439/63504000)*x*y*z^2+(3645611/259200)*y^2*z^2-(572167/181440)*x*z^3-(38930821/3175200)*y*z^3-(1913/900)*z^4-(1726633/680400)*x^3-(1261033/113400)*x^2*y+(2136649/181440)*x*y^2+(240227/7200)*y^3-(22151083/5443200)*x^2*z-(331428401/7257600)*x*y*z-(971543/201600)*y^2*z+(5269/20160)*x*z^2+(2209661/50400)*y*z^2+(5588/1575)*z^3+(82753/9072)*x^2-(12559/151200)*x*y+(16127/1920)*y^2-(13297943/604800)*x*z+(140491/10080)*y*z+(59687/1680)*z^2+(11329/1440)*x-(71/80)*y-(23/80)*z-333/32, -(43/8)*x^6+(36529/5040)*x^5*y+(5549/1890)*x^4*y^2+(432449/264600)*x^3*y^3-(357649/19845)*x^2*y^4-(53753/6750)*x*y^5-(368/375)*y^6+(1454093/141120)*x^5*z-(2570719/529200)*x^4*y*z-(1231037/231525)*x^3*y^2*z+(2643317/1587600)*x^2*y^3*z-(190277/11760)*x*y^4*z-(51/1400)*y^5*z-(534629/28224)*x^4*z^2+(2511667/176400)*x^3*y*z^2-(1103696501/44452800)*x^2*y^2*z^2+(8494807/907200)*x*y^3*z^2+(1069823/31500)*y^4*z^2-(24217/33075)*x^3*z^3-(49530571/2116800)*x^2*y*z^3-(24642463/6350400)*x*y^2*z^3-(126629/75600)*y^3*z^3-(19461481/1058400)*x^2*z^4+(915533/3175200)*x*y*z^4-(12689/1440)*y^2*z^4-(669097/226800)*x*z^5-(601199/56700)*y*z^5-(2057/540)*z^6-(63557/7560)*x^5-(6826709/793800)*x^4*y+(17433953/1058400)*x^3*y^2+(5719663/1190700)*x^2*y^3+(17566163/1134000)*x*y^4+(8767/1125)*y^5-(5298761/1058400)*x^4*z-(677579281/63504000)*x^3*y*z+(399157901/95256000)*x^2*y^2*z+(505720987/31752000)*x*y^3*z+(196961/15120)*y^4*z+(221983/33600)*x^3*z^2+(167343703/7620480)*x^2*y*z^2+(1697257/47628)*x*y^2*z^2+(9103099/151200)*y^3*z^2-(2548711/1905120)*x^2*z^3+(160965253/5080320)*x*y*z^3+(4294153/151200)*y^2*z^3-(7729/36288)*x*z^4+(881/2160)*y*z^4-(3581/1080)*z^5+(1219501/272160)*x^4+(385892197/57153600)*x^3*y+(498801683/9525600)*x^2*y^2+(4714901/453600)*x*y^3+(2053/180)*y^4+(41170111/15240960)*x^3*z-(329042429/10584000)*x^2*y*z+(15214099/31752000)*x*y^2*z+(2161913/45360)*y^3*z+(196918811/8467200)*x^2*z^2+(711856841/15240960)*x*y*z^2+(7240981/181440)*y^2*z^2+(2180833/113400)*x*z^3+(47030719/1814400)*y*z^3+(5569/1080)*z^4+(368041/90720)*x^3-(17094137/2721600)*x^2*y+(20063299/907200)*x*y^2+(7937/360)*y^3-(4969597/907200)*x^2*z-(16176199/1814400)*x*y*z+(53349/11200)*y^2*z+(4098677/181440)*x*z^2+(7060141/181440)*y*z^2+(4529/288)*z^3+(3439/1260)*x^2-(1352627/201600)*x*y+(707/120)*y^2-(3149401/362880)*x*z-(5095/2688)*y*z+(1523/72)*z^2+(839/960)*x-(603/160)*y-(469/64)*z-531/64, -(101/20)*x^6-(126451/12600)*x^5*y-(52649/2100)*x^4*y^2+(53153/31500)*x^3*y^3+(692599/75600)*x^2*y^4+(11197/5400)*x*y^5-(7/36)*y^6+(903/50)*x^5*z+(3502991/75600)*x^4*y*z-(50046163/2268000)*x^3*y^2*z-(680647/33600)*x^2*y^3*z-(17982049/680400)*x*y^4*z-(37271/6480)*y^5*z-(9647/50400)*x^4*z^2+(189563599/4233600)*x^3*y*z^2+(737187221/19051200)*x^2*y^2*z^2-(174213217/6350400)*x*y^3*z^2-(4309649/453600)*y^4*z^2+(723553/50400)*x^3*z^3+(144053197/3528000)*x^2*y*z^3+(117944923/6804000)*x*y^2*z^3-(42815051/9525600)*y^3*z^3+(81071/10080)*x^2*z^4+(6978289/302400)*x*y*z^4+(205894669/47628000)*y^2*z^4+(27813/5600)*x*z^5+(3633277/1058400)*y*z^5+(521/840)*z^6+(19/36)*x^5+(3344137/226800)*x^4*y+(1452887/75600)*x^3*y^2+(35347231/1360800)*x^2*y^3+(5733463/907200)*x*y^4-(10369/4320)*y^5+(101089/2520)*x^4*z+(11028331/317520)*x^3*y*z-(54153679/9525600)*x^2*y^2*z-(438043297/4762800)*x*y^3*z-(66679/2100)*y^4*z+(118071/5600)*x^3*z^2+(1549714217/27216000)*x^2*y*z^2-(16287117607/381024000)*x*y^2*z^2-(613672261/21168000)*y^3*z^2+(3515423/220500)*x^2*z^3-(11743427/15876000)*x*y*z^3-(89419703/10584000)*y^2*z^3+(2229499/470400)*x*z^4+(781651/508032)*y*z^4+(132877/176400)*z^5-(64421/2160)*x^4-(11103277/1360800)*x^3*y+(3832951/453600)*x^2*y^2-(17641/189000)*x*y^3-(18203/14400)*y^4-(2501867/45360)*x^3*z-(260838553/3175200)*x^2*y*z-(5295087649/31752000)*x*y^2*z-(111827287/1512000)*y^3*z-(18942281/362880)*x^2*z^2-(6610196443/50803200)*x*y*z^2-(271403081/5292000)*y^2*z^2-(34141027/846720)*x*z^3-(31854803/1411200)*y*z^3-(3474301/705600)*z^4-(4121/2160)*x^3+(2152697/129600)*x^2*y-(9040487/907200)*x*y^2+(235/72)*y^3-(2816389/75600)*x^2*z-(22377529/254016)*x*y*z-(6072571/100800)*y^2*z-(20856991/604800)*x*z^2-(184951531/5644800)*y*z^2-(4480477/705600)*z^3+(20401/4320)*x^2-(62371/4200)*x*y+(7809/320)*y^2-(1585813/151200)*x*z-(250301/4800)*y*z-(126821/9600)*z^2+(3607/180)*x+(21371/320)*y+(557/112)*z+1071/16, (14/3)*x^6+(2329/420)*x^5*y-(699059/26460)*x^4*y^2+(3691/264600)*x^3*y^3-(899279/132300)*x^2*y^4+(7692/875)*x*y^5+(31/12)*y^6-(6137/420)*x^5*z+(77009/8400)*x^4*y*z+(39870697/3175200)*x^3*y^2*z-(104196787/4762800)*x^2*y^3*z+(22768199/3969000)*x*y^4*z-(18019/1800)*y^5*z-(546629/100800)*x^4*z^2-(740849/90720)*x^3*y*z^2-(59576507/9525600)*x^2*y^2*z^2+(10123541/1058400)*x*y^3*z^2-(68791/10800)*y^4*z^2-(4633133/302400)*x^3*z^3-(74929/33600)*x^2*y*z^3+(79300723/9525600)*x*y^2*z^3-(69497/20160)*y^3*z^3-(231319/56700)*x^2*z^4-(531493/100800)*x*y*z^4+(139271/90720)*y^2*z^4-(9563/3360)*x*z^5+(5479/12096)*y*z^5+(391/2160)*z^6+(1853/360)*x^5-(129821/113400)*x^4*y-(5434711/1587600)*x^3*y^2+(206561/64800)*x^2*y^3+(101627/3780)*x*y^4+(5923/600)*y^5+(3780493/226800)*x^4*z+(1125769457/23814000)*x^3*y*z+(234841721/83349000)*x^2*y^2*z+(4883177/226800)*x*y^3*z-(3355691/151200)*y^4*z+(11462071/3175200)*x^3*z^2+(22573379/604800)*x^2*y*z^2+(567793459/13891500)*x*y^2*z^2-(22574443/1587600)*y^3*z^2+(31502839/6350400)*x^2*z^3+(709098209/38102400)*x*y*z^3-(10823321/6350400)*y^2*z^3+(293119/635040)*x*z^4+(46029653/5080320)*y*z^4+(2951/1728)*z^5-(767033/32400)*x^4+(86393/75600)*x^3*y-(52617421/1984500)*x^2*y^2+(1785493/45360)*x*y^3+(11999/720)*y^4+(284311/16200)*x^3*z+(53483/3175200)*x^2*y*z+(184785661/11907000)*x*y^2*z-(10881973/151200)*y^3*z-(362885699/25401600)*x^2*z^2+(46670083/2116800)*x*y*z^2-(67828421/2116800)*y^2*z^2-(2004203/25401600)*x*z^3-(103485733/5080320)*y*z^3-(134527/30240)*z^4+(148793/12960)*x^3+(1782961/113400)*x^2*y+(1557827/30240)*x*y^2+(10981/480)*y^3+(44419097/2721600)*x^2*z+(246738319/6350400)*x*y*z-(137771/2240)*y^2*z+(968323/396900)*x*z^2-(2361599/84672)*y*z^2-(224557/20160)*z^3+(1199449/43200)*x^2+(128017/2240)*x*y+(1739/80)*y^2+(34483621/907200)*x*z-(1129/63)*y*z-(214547/40320)*z^2+(2599/60)*x+(1035/32)*y+(1161/80)*z+639/16, (15/4)*x^6-(7811/5880)*x^5*y+(90403/29400)*x^4*y^2+(180401/11025)*x^3*y^3-(572491/12600)*x^2*y^4+(694333/37800)*x*y^5-(469/90)*y^6-(1399/1120)*x^5*z-(310577/117600)*x^4*y*z-(93897719/2646000)*x^3*y^2*z+(8777939/1764000)*x^2*y^3*z+(575129/21600)*x*y^4*z-(205259/4200)*y^5*z+(301643/50400)*x^4*z^2+(48892999/10584000)*x^3*y*z^2-(15110183/882000)*x^2*y^2*z^2-(65666327/3969000)*x*y^3*z^2-(172567/6000)*y^4*z^2+(8803/2520)*x^3*z^3+(58366387/3528000)*x^2*y*z^3+(109429927/15876000)*x*y^2*z^3-(1029331/302400)*y^3*z^3+(96953/12600)*x^2*z^4+(91482443/6350400)*x*y*z^4+(19795369/2268000)*y^2*z^4+(48443/12600)*x*z^5+(2391707/302400)*y*z^5+(37/24)*z^6+(1343/140)*x^5+(24557/5880)*x^4*y+(485399/75600)*x^3*y^2+(416237/21000)*x^2*y^3+(43973123/2268000)*x*y^4-(90113/5400)*y^5+(383909/58800)*x^4*z+(9587758951/222264000)*x^3*y*z+(4095697/216000)*x^2*y^2*z+(555196927/10584000)*x*y^3*z-(18948401/756000)*y^4*z+(369821/25200)*x^3*z^2+(4701868229/333396000)*x^2*y*z^2+(39287599/3969000)*x*y^2*z^2-(150216781/3969000)*y^3*z^2+(36475/6048)*x^2*z^3+(3551689/181440)*x*y*z^3-(34806349/1764000)*y^2*z^3+(60101/13230)*x*z^4-(19839887/2822400)*y*z^4+(5869/8400)*z^5+(1846717/226800)*x^4-(23974597/3175200)*x^3*y-(690342211/6804000)*x^2*y^2-(13130807/1134000)*x*y^3-(185219/3600)*y^4-(116834413/12700800)*x^3*z-(175142299/12700800)*x^2*y*z-(28534591/756000)*x*y^2*z-(94226747/756000)*y^3*z-(41734883/4233600)*x^2*z^2-(309684919/8467200)*x*y*z^2-(4786829/37800)*y^2*z^2-(2765183/264600)*x*z^3-(22368259/564480)*y*z^3-(100873/11200)*z^4+(1267871/136080)*x^3+(128353/2800)*x^2*y-(386551/18900)*x*y^2-(971/240)*y^3-(40316993/3175200)*x^2*z+(43529203/2116800)*x*y*z-(1349153/50400)*y^2*z-(58790633/4233600)*x*z^2-(54765671/1058400)*y*z^2-(421891/33600)*z^3+(4559393/226800)*x^2+(3208603/151200)*x*y+(593/240)*y^2+(104519/12600)*x*z-(26287/16800)*y*z-(88687/13440)*z^2+(257/120)*x+(2153/80)*y+(5217/560)*z+171/16];

        result = solve(f)
        @test nsolutions(result) == 80
    end
end
